import pandas as pd

from aistac.handlers.abstract_handlers import ConnectorContract
from aistac.components.abstract_component import AbstractComponent
from ds_discovery.transition.discovery import DataDiscovery

from ds_models.intent.models_intent_model import ModelsIntentModel
from ds_models.managers.models_property_manager import ModelsPropertyManager

__author__ = 'Darryl Oatridge'


class ModelsComponent(AbstractComponent):

    CONNECTOR_FEATURE = 'feature_connector'
    CONNECTOR_TRAINED = 'trained_connector'
    CONNECTOR_PREDICT = 'predict_connector'

    def __init__(self, property_manager: ModelsPropertyManager, intent_model: ModelsIntentModel,
                 default_save=None):
        """ Encapsulation class for the discovery set of classes

        :param property_manager: The contract property manager instance for this component
        :param intent_model: the model codebase containing the parameterizable intent
        :param default_save: The default behaviour of persisting the contracts:

                    if False: The connector contracts are kept in memory (useful for restricted file systems)
        """
        super().__init__(property_manager=property_manager, intent_model=intent_model, default_save=default_save)

    @classmethod
    def from_uri(cls, task_name: str, uri_pm_path: str, default_save=None, **kwargs):
        """ Class Factory Method to instantiates the component application. The Factory Method handles the
        instantiation of the Properties Manager, the Intent Model and the persistence of the uploaded properties.

        by default the handler is local Pandas but also supports remote AWS S3 and Redis. It use these Factory
        instantiations ensure that the schema is s3:// or redis:// and the handler will be automatically redirected

         :param task_name: The reference name that uniquely identifies a task or subset of the property manager
         :param uri_pm_path: A URI that identifies the resource path for the property manager.
         :param default_save: (optional) if the configuration should be persisted. default to 'True'
         :param kwargs: to pass to the connector contract
         :return: the initialised class instance
         """
        _pm = ModelsPropertyManager(task_name=task_name)
        _intent_model = ModelsIntentModel(property_manager=_pm)
        super()._init_properties(property_manager=_pm, uri_pm_path=uri_pm_path, **kwargs)
        return cls(property_manager=_pm, intent_model=_intent_model, default_save=default_save)

    @classmethod
    def _from_remote_s3(cls) -> (str, str):
        """ Class Factory Method that builds the connector handlers an Amazon AWS s3 remote store."""
        _module_name = 'ds_connectors.handlers.aws_s3_handlers'
        _handler = 'AwsS3PersistHandler'
        return _module_name, _handler

    @classmethod
    def _from_remote_redis(cls) -> (str, str):
        """ Class Factory Method that builds the connector handlers an Amazon AWS s3 remote store."""
        _module_name = 'ds_connectors.handlers.redis_handlers'
        _handler = 'RedisPersistHandler'
        return _module_name, _handler

    @staticmethod
    def canonical_report(df, stylise: bool=True, inc_next_dom: bool=False, report_header: str=None,
                         condition: str=None):
        """The Canonical Report is a data dictionary of the canonical providing a reference view of the dataset's
        attribute properties

        :param df: the DataFrame to view
        :param stylise: if True present the report stylised.
        :param inc_next_dom: (optional) if to include the next dominate element column
        :param report_header: (optional) filter on a header where the condition is true. Condition must exist
        :param condition: (optional) the condition to apply to the header. Header must exist. examples:
                ' > 0.95', ".str.contains('shed')"
        :return:
        """
        return DataDiscovery.data_dictionary(df=df, stylise=stylise, inc_next_dom=inc_next_dom,
                                             report_header=report_header, condition=condition)

    def is_feature_frame_modified(self):
        """Test if the source file is modified since last load"""
        return self.is_canonical_modified(self.CONNECTOR_FEATURE)

    def set_feature_catalog_connector(self, connector_contract: ConnectorContract, save: bool=None):
        """ Sets the feature catalog connector contract

        :param connector_contract: a Connector Contract for the properties persistence
        :param save: (optional) if True, save to file. Default is True
        """
        save = save if isinstance(save, bool) else self._default_save
        if self.has_connector_contract(self.CONNECTOR_FEATURE):
            self.remove_connector_contract(self.CONNECTOR_FEATURE)
        self.set_connector_contract(self.CONNECTOR_FEATURE, connector_contract=connector_contract, save=save)
        self.pm_persist(save)
        return

    def set_trained_connector(self, connector_contract: ConnectorContract, save: bool=None):
        """ Sets the trained model connector contract

        :param connector_contract: a Connector Contract for the properties persistence
        :param save: (optional) if True, save to file. Default is True
        """
        save = save if isinstance(save, bool) else self._default_save
        if self.has_connector_contract(self.CONNECTOR_TRAINED):
            self.remove_connector_contract(self.CONNECTOR_TRAINED)
        self.set_connector_contract(self.CONNECTOR_TRAINED, connector_contract=connector_contract, save=save)
        self.pm_persist(save)
        return

    def set_predict_connector(self, connector_contract: ConnectorContract, save: bool=None):
        """ Sets the predict connector contract

        :param connector_contract: a Connector Contract for the properties persistence
        :param save: (optional) if True, save to file. Default is True
        """
        save = save if isinstance(save, bool) else self._default_save
        if self.has_connector_contract(self.CONNECTOR_PREDICT):
            self.remove_connector_contract(self.CONNECTOR_PREDICT)
        self.set_connector_contract(self.CONNECTOR_PREDICT, connector_contract=connector_contract, save=save)
        self.pm_persist(save)
        return

    def load_feature_catalog_canonical(self) -> pd.DataFrame:
        """returns the feature catalog as a DataFrame """
        return self.load_canonical(self.CONNECTOR_FEATURE)

    def load_trained_canonical(self) -> pd.DataFrame:
        """returns the model as a DataFrame """
        return self.load_canonical(self.CONNECTOR_TRAINED)

    def load_predict_canonical(self) -> pd.DataFrame:
        """returns the predict as a DataFrame """
        return self.load_canonical(self.CONNECTOR_PREDICT)

    def load_canonical(self, connector_name: str) -> pd.DataFrame:
        """returns the canonical of the referenced connector

        :param connector_name: the name or label to identify and reference the connector
        """
        canonical = super().load_canonical(connector_name=connector_name)
        if isinstance(canonical, dict):
            canonical = pd.DataFrame.from_dict(data=canonical, orient='columns')
        return canonical

    def save_trained_canonical(self, df):
        """Saves the pandas.DataFrame to the clean files folder"""
        self.persist_canonical(self.CONNECTOR_TRAINED, df)

    def save_predict_canonical(self, df):
        """Saves the pandas.DataFrame to the clean files folder"""
        self.persist_canonical(self.CONNECTOR_PREDICT, df)

    def report_intent(self, stylise: bool=True):
        """ generates a report on all the intent

        :param stylise: returns a stylised dataframe with formatting
        :return: pd.Dataframe
        """
        stylise = True if not isinstance(stylise, bool) else stylise
        style = [{'selector': 'th', 'props': [('font-size', "120%"), ("text-align", "center")]},
                 {'selector': '.row_heading, .blank', 'props': [('display', 'none;')]}]
        df = pd.DataFrame.from_dict(data=self.pm.report_intent(), orient='columns')
        if stylise:
            index = df[df['level'].duplicated()].index.to_list()
            df.loc[index, 'level'] = ''
            df = df.reset_index(drop=True)
            df_style = df.style.set_table_styles(style).set_properties(**{'text-align': 'left'})
            _ = df_style.set_properties(subset=['level'],  **{'font-weight': 'bold', 'font-size': "120%"})
            return df_style
        return df

    def report_connectors(self, connector_filter: [str, list]=None, stylise: bool=True):
        """ generates a report on the source contract

        :param connector_filter: (optional) filters on the connector name.
        :param stylise: (optional) returns a stylised DataFrame with formatting
        :return: pd.DataFrame
        """
        stylise = True if not isinstance(stylise, bool) else stylise
        style = [{'selector': 'th', 'props': [('font-size', "120%"), ("text-align", "center")]},
                 {'selector': '.row_heading, .blank', 'props': [('display', 'none;')]}]
        df = pd.DataFrame.from_dict(data=super().report_connectors(connector_filter=connector_filter), orient='columns')
        if stylise:
            df_style = df.style.set_table_styles(style).set_properties(**{'text-align': 'left'})
            _ = df_style.set_properties(subset=['connector_name'], **{'font-weight': 'bold'})
            return df_style
        else:
            df.set_index(keys='connector_name', inplace=True)
        return df

    def report_notes(self, catalog: [str, list]=None, labels: [str, list]=None, regex: [str, list]=None,
                     re_ignore_case: bool=False, stylise: bool=True, drop_dates: bool=False):
        """ generates a report on the notes

        :param catalog: (optional) the catalog to filter on
        :param labels: (optional) s label or list of labels to filter on
        :param regex: (optional) a regular expression on the notes
        :param re_ignore_case: (optional) if the regular expression should be case sensitive
        :param stylise: (optional) returns a stylised dataframe with formatting
        :param drop_dates: (optional) excludes the 'date' column from the report
        :return: pd.Dataframe
        """
        stylise = True if not isinstance(stylise, bool) else stylise
        drop_dates = False if not isinstance(drop_dates, bool) else drop_dates
        style = [{'selector': 'th', 'props': [('font-size', "120%"), ("text-align", "center")]},
                 {'selector': '.row_heading, .blank', 'props': [('display', 'none;')]}]
        report = super().report_notes(catalog=catalog, labels=labels, regex=regex, re_ignore_case=re_ignore_case,
                                      drop_dates=drop_dates)
        df = pd.DataFrame.from_dict(data=report, orient='columns')
        if stylise:
            df_style = df.style.set_table_styles(style).set_properties(**{'text-align': 'left'})
            _ = df_style.set_properties(subset=['section'], **{'font-weight': 'bold'})
            _ = df_style.set_properties(subset=['label', 'section'], **{'font-size': "120%"})
            return df_style
        return df
