import inspect
import threading
from copy import deepcopy
from typing import Any

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
import sklearn.metrics as skmetrics
from catboost import CatBoostClassifier

from aistac.properties.abstract_properties import AbstractPropertyManager
from aistac.intent.abstract_intent import AbstractIntentModel
from ds_discovery.intent.transition_intent import TransitionIntentModel
from ds_models.models.claims_code_categorizer import categorizer

__author__ = 'Darryl Oatridge'


class ModelsIntentModel(AbstractIntentModel):

    TRAIN_INTENT_LEVEL = 'train_level'
    PREDICT_INTENT_LEVEL = 'predict_level'

    def run_intent_pipeline(self, level: [int, str, list]=None, **kwargs):
        """ Collectively runs all parameterised intent taken from the property manager against the code base as
        defined by the intent_contract.

        It is expected that all intent methods have the 'canonical' as the first parameter of the method signature
        and will contain 'inplace' and 'save_intent' as parameters.

        :param canonical: this is the iterative value all intent are applied to and returned.
        :param levels: (optional) an single or list of levels to run, if list, run in order given
        :param inplace: (optional) change data in place or to return a deep copy. default False
        :param kwargs: additional kwargs to add to the parameterised intent, these will replace any that already exist
        :return Canonical with parameterised intent applied or None if inplace is True
        """
        # test if there is any intent to run
        if self._pm.has_intent():
            # get the list of levels to run
            if level == self.TRAIN_INTENT_LEVEL:
                for method, params in self._pm.get_intent(level=level).items():
                    if method in self.__dir__():
                        feature_catalog = kwargs.get('feature_catalog', pd.DataFrame())
                        return eval(f"self.{method}(feature_catalog={feature_catalog}, save_intent=False, **{params})")
        return None

    def __init__(self, property_manager: AbstractPropertyManager, default_save_intent: bool=True,
                 intent_next_available: bool=False):
        """initialisation of the Intent class. The 'intent_param_exclude' is used to exclude commonly used method
         parameters from being included in the intent contract, this is particularly useful if passing a canonical, or
         non relevant parameters to an intent method pattern. Any named parameter in the intent_param_exclude list
         will not be included in the recorded intent contract for that method

        :param property_manager: the property manager class that references the intent contract.
        :param default_save_intent: (optional) The default action for saving intent in the property manager
        :param intent_next_available: (optional) if the default level should be set to next available level or zero
        """
        default_save_intent = default_save_intent if isinstance(default_save_intent, bool) else True
        intent_param_exclude = ['inplace', 'model']
        super().__init__(property_manager=property_manager, intent_param_exclude=intent_param_exclude,
                         default_save_intent=default_save_intent)
        # globals
        self._default_intent_level = -1 if isinstance(intent_next_available, bool) and intent_next_available else 0

    def cat_boost_classifier_train(self, feature_frame: pd.DataFrame, target: str, headers: list, run_hyperopt: bool=True,
                                test_size: float=None, random_state: int=None, save_intent: bool=True,
                                **cat_boost_kwargs):
        """ trains a Claims Classifier

        :param feature_frame: the model feature frame in the format accepted by the model
        :param target: the target name
        :param headers: the cat header names
        :param run_hyperopt: (optional) if Hyperopt should be used (default true)
        :param test_size:  (optional) the size of the test percentage (default tp 0.3)
        :param random_state:  (optional) a random state value to return the same selection each run
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param cat_boost_kwargs: kwargs to pass to the CatBoostClassifier
        :return: CatBoostClassifier.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals(),
                                                        exclude=['feature_frame']),
                                   intent_level=self.TRAIN_INTENT_LEVEL, save_intent=save_intent)
        # Code block for intent
        X = self.filter_columns(feature_frame, headers=headers)
        y = self.filter_columns(feature_frame, headers=target)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=random_state)
        model = CatBoostClassifier(**cat_boost_kwargs)
        model.fit(X_train, y_train, cat_features=headers, eval_set=[(X_test, y_test)])
        return model

    def cat_boost_classifier_predict(self, model: Any, feature_frame: pd.DataFrame, headers: list, prediction_type=None,
                                  save_intent: bool=True):
        """ Predicts

        :param model: the CatBoostClassifier
        :param feature_frame: the model feature frame in the format accepted by the model
        :param headers: the cat header names
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level of the intent,
        :return: value prediction.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals(),
                                                        exclude=['model', 'feature_frame']),
                                   intent_level=self.PREDICT_INTENT_LEVEL, save_intent=save_intent)
        # Code block for intent
        df_indices = feature_frame.index.values
        feature_frame.reset_index(drop=True, inplace=True)
        X = self.filter_columns(feature_frame, headers=headers)
        for col in headers:
            if col not in feature_frame.columns:
                feature_frame[col] = pd.Series(dtype=int)
        if isinstance(prediction_type, str):
            predict =  model.predict(X, prediction_type=prediction_type)
        else:
            predict =  model.predict_proba(X)
        return pd.DataFrame(predict,index=df_indices)

    @staticmethod
    def predict_and_score(model, feature_frame: pd.DataFrame, target: str, headers: list, prediction_type: str=None):
        X = ModelsIntentModel.filter_columns(feature_frame, headers=headers)
        y = ModelsIntentModel.filter_columns(feature_frame, headers=target)
        if isinstance(prediction_type, str):
            predict = model.predict(X, prediction_type=prediction_type)
        else:
            predict = model.predict_proba(X)

        auc = skmetrics.roc_auc_score(y.astype(int), predict[:, 1])
        f1 = skmetrics.f1_score(y.astype(int), np.array(predict[:, 1] > 0.5).astype(int))
        return auc, f1

    def categorize_procedure_codes(self, df: pd.DataFrame, target_header: str, code_header: str, save_intent: bool=True,
                                   intent_level: [int, str]=None):
        """A set of python functions for categorizing HCPC codes and DRG codes

        :param df:
        :param target_header:
        :param code_header:
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level of the intent,
        :return: modified DataFrame with the aditional column
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals(),
                                                        exclude=['df']),
                                   intent_level=intent_level, save_intent=save_intent)
        # Code block for intent

        df[target_header] = df[code_header].apply(
            lambda x: categorizer.carrier_categorizer_by_hcpc(x))
        return df

    @staticmethod
    def filter_headers(df: pd.DataFrame, headers: [str, list]=None, drop: bool=None, dtype: [str, list]=None,
                       exclude: bool=None, regex: [str, list]=None, re_ignore_case: bool=None) -> list:
        """ returns a list of headers based on the filter criteria

        :param df: the Pandas.DataFrame to get the column headers from
        :param headers: a list of headers to drop or filter on type
        :param drop: to drop or not drop the headers
        :param dtype: the column types to include or exclude. Default None else int, float, bool, object, 'number'
        :param exclude: to exclude or include the dtypes. Default is False
        :param regex: a regular expression to search the headers. example '^((?!_amt).)*$)' excludes '_amt' headers
        :param re_ignore_case: true if the regex should ignore case. Default is False
        :return: a filtered list of headers

        :raise: TypeError if any of the types are not as expected
        """
        return TransitionIntentModel.filter_headers(df, headers=headers, drop=drop, dtype=dtype, exclude=exclude,
                                                    regex=regex, re_ignore_case=re_ignore_case)

    @staticmethod
    def filter_columns(df, headers=None, drop=False, dtype=None, exclude=False, regex=None, re_ignore_case=None,
                       inplace=False) -> [dict, pd.DataFrame]:
        """ Returns a subset of columns based on the filter criteria

        :param df: the Pandas.DataFrame to get the column headers from
        :param headers: a list of headers to drop or filter on type
        :param drop: to drop or not drop the headers
        :param dtype: the column types to include or excluse. Default None else int, float, bool, object, 'number'
        :param exclude: to exclude or include the dtypes
        :param regex: a regular expression to search the headers. example '^((?!_amt).)*$)' excludes '_amt' columns
        :param re_ignore_case: true if the regex should ignore case. Default is False
        :param inplace: if the passed pandas.DataFrame should be used or a deep copy
        :return:
        """
        return TransitionIntentModel.filter_columns(df, headers=headers, drop=drop, dtype=dtype, exclude=exclude,
                                                    regex=regex, re_ignore_case=re_ignore_case, inplace=inplace)

